#!/usr/bin/env python

"""
Simplified program utilities.

Copyright (C) 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from compiler.visitor import ASTVisitor

# Exceptions.

class SimplifiedError(Exception):

    "An error in the annotation process."

    def __init__(self, exc, node, *args):

        """
        Initialise the error with an existing exception 'exc', the 'node' at
        which this error occurs, along with additional optional arguments.
        """

        Exception.__init__(self, *args)
        self.nodes = [node]
        self.exc = exc

    def add(self, node):

        "Add the given 'node' to the path of nodes leading from the exception."

        self.nodes.append(node)

    def __str__(self):

        "Return a string showing the principal exception details."

        return "%s, %s" % (self.exc, self.nodes)

# Elementary visitor support.

class Visitor(ASTVisitor):

    "A visitor base class."

    def __init__(self):
        ASTVisitor.__init__(self)

    def default(self, node, *args):
        raise SimplifiedError, (None, node)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def dispatches(self, nodes, *args):
        results = []
        for node in nodes:
            results.append(self.dispatch(node, *args))
        return results

    def dispatch_dict(self, d, *args):
        results = {}
        for name, node in d.items():
            results[name] = self.dispatch(node, *args)
        return results

# Unique name registration.

class Naming:

    "Maintain records of unique names for each simple name."

    index_separator = "-"

    def __init__(self):
        self.names = {}

    def get(self, obj):
        return obj._unique_name

    def set(self, obj, name):
        if hasattr(obj, "_unique_name"):
            return
        if not self.names.has_key(name):
            self.names[name] = 0
        n = self.names[name] + 1
        self.names[name] = n
        obj._unique_name = "%s%s%d" % (name, self.index_separator, n)

def name(obj, name):

    "Return a unique name for the given 'obj', indicating the base 'name'."

    naming.set(obj, name)
    return naming.get(obj)

# Naming singleton.

naming = Naming()

# Named nodes are those which can be referenced in some way.

class WithName:

    "Node naming."

    def __init__(self):

        "Initialise the object's full name."

        self._full_name = name(self, self.name or "$untitled")

    def full_name(self):

        "Return the object's full name."

        return self._full_name

# Comparable nodes based on naming.

class Comparable:

    "Comparable nodes implementing the 'full_name' method."

    def __eq__(self, other):

        "This object is equal to 'other' if the full names are the same."

        # NOTE: Single instance: all instances are the same
        # NOTE: Multiple instances: all instances are different
        if hasattr(other, "full_name"):
            return self.full_name() == other.full_name()
        else:
            return NotImplemented

    def __hash__(self):

        "The hash of this object is based on its full name."

        return hash(self.full_name())

# Structure nodes indicating namespace-bearing objects.

class Structure(Comparable):

    "A non-program node containing some kind of namespace."

    def __init__(self, **kw):
        for name, value in kw.items():
            setattr(self, name, value)

    def __repr__(self):
        return "%s '%s'" % (self.__class__.__name__, self.full_name())

# vim: tabstop=4 expandtab shiftwidth=4
